#![no_std]
#![no_main]

use panic_halt as _;

#[rtic::app(device = waveshare_rp2040_zero::hal::pac, peripherals = true)]
mod app {
    use core::iter::once;use fugit::MicrosDurationU32;
    use rp2040_hal::Clock;
    use rp2040_hal::gpio::{FunctionPio0, PullDown};
    use rp2040_hal::gpio::bank0::Gpio16;
    use rp2040_hal::pac::PIO0;
    use rp2040_hal::pio::SM0;
    use rp2040_hal::timer::CountDown;
    use smart_leds::{brightness, SmartLedsWrite};
    use usbd_serial::SerialPort;
    use waveshare_rp2040_zero::{
        hal::{self, clocks::init_clocks_and_plls, timer::Alarm, watchdog::Watchdog, Sio},
        XOSC_CRYSTAL_FREQ,
    };

    use usb_device::{class_prelude::*, prelude::*};
    use ws2812_pio::Ws2812;
    use waveshare_rp2040_zero::hal::pio::PIOExt;

    const SCAN_TIME_US: MicrosDurationU32 = MicrosDurationU32::millis(400);

    #[shared]
    struct Shared {
        timer: hal::Timer,
        alarm: hal::timer::Alarm0,
        usb_dev: UsbDevice<'static, hal::usb::UsbBus>,
        serial: SerialPort<'static, hal::usb::UsbBus>,
        ws: Ws2812<PIO0, SM0, CountDown<'static>, hal::gpio::Pin<Gpio16, FunctionPio0, PullDown>>
    }

    #[local]
    struct Local {}

    #[init]
    fn init(c: init::Context) -> (Shared, Local, init::Monotonics) {
        // Soft-reset does not release the hardware spinlocks
        // Release them now to avoid a deadlock after debug or watchdog reset
        unsafe {
            hal::sio::spinlock_reset();
        }
        let mut resets = c.device.RESETS;
        let mut watchdog = Watchdog::new(c.device.WATCHDOG);
        let clocks = init_clocks_and_plls(
            XOSC_CRYSTAL_FREQ,
            c.device.XOSC,
            c.device.CLOCKS,
            c.device.PLL_SYS,
            c.device.PLL_USB,
            &mut resets,
            &mut watchdog,
        )
            .ok()
            .unwrap();

        let sio = Sio::new(c.device.SIO);
        let pins = waveshare_rp2040_zero::Pins::new(
            c.device.IO_BANK0,
            c.device.PADS_BANK0,
            sio.gpio_bank0,
            &mut resets,
        );

        //let timer = Timer::new(c.device.TIMER, &mut c.device.RESETS, &clocks);
        let dTIMER = c.device.TIMER;
        let timer = cortex_m::singleton!(: hal::Timer = hal::Timer::new(dTIMER, &mut resets, &clocks)).unwrap();
        let mut alarm = timer.alarm_0().unwrap();

            let count_down: CountDown<'static> = timer.count_down();

            let (mut pio, sm0, _, _, _) = c.device.PIO0.split(&mut resets);
            let ws: Ws2812<PIO0, SM0, CountDown<'static>, hal::gpio::Pin<Gpio16, FunctionPio0, PullDown>> = Ws2812::new(
                // The onboard NeoPixel is attached to GPIO pin #16 on the Waveshare RP2040-Zero.
                pins.neopixel.into_function(),
                &mut pio,
                sm0,
                clocks.peripheral_clock.freq(),
                count_down,
            );








        // Set up the USB driver
        let USBCTRL_REGS = c.device.USBCTRL_REGS;
        let USBCTRL_DPRAM = c.device.USBCTRL_DPRAM;
        let usb_clock = clocks.usb_clock;
        let usb_bus = cortex_m::singleton!(
            : UsbBusAllocator<hal::usb::UsbBus> = UsbBusAllocator::new(waveshare_rp2040_zero::hal::usb::UsbBus::new(
            USBCTRL_REGS,
            USBCTRL_DPRAM,
            usb_clock,
            true,
            &mut resets,
        ))).unwrap();

        // Set up the USB Communications Class Device driver
        let mut serial = SerialPort::new(usb_bus);

        // Create a USB device with a fake VID and PID
        let usb_dev = UsbDeviceBuilder::new(usb_bus, UsbVidPid(0x16c0, 0x27dd))
            .strings(&[StringDescriptors::default()
                .manufacturer("Fake company")
                .product("Serial port")
                .serial_number("TEST")])
            .unwrap()
            .device_class(2) // from: https://www.usb.org/defined-class-codes
            .build();

        let _ = serial.write(b"Hello, World!\r\n");



        let _ = alarm.schedule(SCAN_TIME_US);
        alarm.enable_interrupt();

        (Shared { timer: *timer, alarm, usb_dev, serial, ws}, Local {}, init::Monotonics())
    }

    #[task(
        binds = TIMER_IRQ_0,
        priority = 1,
        shared = [alarm, serial, ws],
        local = [tog: bool = true],
    )]
    fn timer_irq(mut c: timer_irq::Context) {
        if *c.local.tog {
            c.shared.ws.lock(|ws| ws.write(brightness(once((255, 0, 0).into()), 32)).unwrap());
            c.shared.serial.lock(|serial| _ = serial.write(b"High\r\n"));

        } else {
            c.shared.ws.lock(|ws| ws.write(brightness(once((0, 0, 0).into()), 32)).unwrap());
            c.shared.serial.lock(|serial| _ = serial.write(b"LOW\r\n"));
        }
        *c.local.tog = !*c.local.tog;


        let mut alarm = c.shared.alarm;
        (alarm).lock(|a| {
            a.clear_interrupt();
            let _ = a.schedule(SCAN_TIME_US);
        });
    }

    #[task(
        binds = USBCTRL_IRQ,
        priority = 1,
        shared = [usb_dev, serial],
    )]
    fn usbctrl_irq(c: usbctrl_irq::Context) {
        // Grab the global objects. This is OK as we only access them under interrupt.
        let usb_dev = c.shared.usb_dev;
        let serial = c.shared.serial;

        (usb_dev, serial).lock(|usb_dev, serial| {
            // Poll the USB driver with all of our supported USB Classes

            if usb_dev.poll(&mut [serial]) {
                let mut buf = [0u8; 64];
                match serial.read(&mut buf) {
                    Err(_e) => {
                        // Do nothing
                    }
                    Ok(0) => {
                        // Do nothing
                    }
                    Ok(count) => {
                        // Convert to upper case
                        buf.iter_mut().take(count).for_each(|b| {
                            b.make_ascii_uppercase();
                        });

                        // Send back to the host
                        let mut wr_ptr = &buf[..count];
                        while !wr_ptr.is_empty() {
                            let _ = serial.write(wr_ptr).map(|len| {
                                wr_ptr = &wr_ptr[len..];
                            });
                        }
                    }
                }
            }
        });

    }
}